/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Calulate the face centres and areas.

    Calculate the centre by breaking the face into triangles using the face
    centre and area-weighted averaging their centres.  This method copes with
    small face-concavity.
    
    For spherical geometry - only copes with either purely radail or purely spherical
    faces - not a mixture

\*---------------------------------------------------------------------------*/

#include "primitiveMesh.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void primitiveMesh::calcFaceCentresAndAreas() const
{
    //if (debug)
    {
        Pout<< "primitiveMesh::calcFaceCentresAndAreas() : "
            << "Calculating face centres and face areas"
            << " for sperical geometry"
            << endl;
    }

    // It is an error to attempt to recalculate faceCentres
    // if the pointer is already set
    if (faceCentresPtr_ || faceAreasPtr_)
    {
        FatalErrorIn("primitiveMesh::calcFaceCentresAndAreas() const")
            << "Face centres or face areas already calculated"
            << exit(FatalError);
    }

    faceCentresPtr_ = new vectorField(nFaces());
    vectorField& fCtrs = *faceCentresPtr_;

    faceAreasPtr_ = new vectorField(nFaces());
    vectorField& fAreas = *faceAreasPtr_;

    makeFaceCentresAndAreas(points(), fCtrs, fAreas);

    if (debug)
    {
        Pout<< "primitiveMesh::calcFaceCentresAndAreas() : "
            << "Finished calculating face centres and face areas"
            << endl;
    }
}


void primitiveMesh::makeFaceCentresAndAreas
(
    const pointField& p,
    vectorField& fCtrs,
    vectorField& fAreas
) const
{
    const faceList& fs = faces();

    forAll (fs, facei)
    {
        const labelList& f = fs[facei];
        label nPoints = f.size();
        
        // reference radii for the face
        scalar r = mag(p[f[0]]);
        scalar rS = sqr(r);

        scalar r1 = r;
        label ipr1 = 0;
        for(label ip = 1; ip < nPoints && mag(r1 - r) < SMALL*r; ip++)
        {
            r1 = mag(p[f[ip]]);
            ipr1 = ip;
        }

        // Determine whether it is a radial or spherical face
        
        // for a radial face - 4 vertices with 2 different radii
        if(nPoints == 4 && mag(r1 - r) > r*SMALL)
        {
            // check that there are only 2 different radii
            for(label ip = 1; ip < 4; ip++)
            {
                if(mag(r - mag(p[f[ip]])) > r*SMALL 
                    && mag(r1 - mag(p[f[ip]])) > r1*SMALL)
                {
                    FatalErrorIn("primitiveMesh::makeFaceCentresAndAreas")
                        << "face " << facei << " has 3 different radii, "
                        << r << ", " << r1 << " and " << mag(p[f[ip]]) << endl
                        << exit(FatalError);
                }   
            }
            
            // Find the angle swept out by the face and the vertex on the other side
            scalar theta = 0;
            label ipt = 0;
            for(label ip = 1; ip < 4 && mag(theta) < SMALL; ip++)
            {
                if(mag(r - mag(p[f[ip]])) < r*SMALL)
                {
                    theta = acos((p[f[0]] & p[f[ip]])/rS);
                    ipt = ip;
                }
            }
            
            // face centre
            fCtrs[facei] = p[f[0]] + p[f[ipt]];
            fCtrs[facei] = fCtrs[facei]/mag(fCtrs[facei])*
            (
//                (2./3.)*sin(theta)/theta*(rS + r*r1 + sqr(r1))/(r+r1)
                (2./3.)*(rS + r*r1 + sqr(r1))/(r+r1)
            );
            
            // area vector
            fAreas[facei] = (p[f[ipt]] - p[f[0]])^fCtrs[facei];
            if(ipt == 1) fAreas[facei] = -fAreas[facei];
            fAreas[facei] = 0.5*fAreas[facei]/mag(fAreas[facei])
                            *theta*(rS - sqr(r1));
        }
        
        // for a spherical face
        else
        {            
            // first find approximate face centre
            point fCentre = vector::zero;
            scalar rCentre = 0.0;
            for (label pi = 0; pi < nPoints; pi++)
            {
                fCentre += p[f[pi]];
                rCentre += mag(p[f[pi]]);
            }

            rCentre /= scalar(nPoints);
            fCentre = fCentre*rCentre/mag(fCentre);

            // next find triangle area (At) and triangle centre (Ct),
            // and sum to get face centre, Cf and sum of face areas, sumAt
            
            vector Cf = vector::zero;
            scalar sumAt = 0;

            for (label pi = 0; pi < nPoints; pi++)
            {
                const point& nextPoint = p[f[(pi + 1) % nPoints]];
                
                scalar At = 2*rS*atan
                (
                    p[f[pi]] & (nextPoint ^ fCentre)/(r*
                    (
                        rS + (nextPoint & fCentre) + (nextPoint & p[f[pi]]) 
                           + (p[f[pi]] & fCentre)
                    ))
                );
                

                vector Ct = p[f[pi]] + nextPoint + fCentre;
                Ct *= (r/mag(Ct));
                
                sumAt += At;;
                Cf += mag(At)*Ct;
            }

            fCtrs[facei] = Cf/mag(Cf);
            fAreas[facei] = fCtrs[facei]*sumAt;
            fCtrs[facei] *= r;
        }
/*
        if(facei < 10 || (facei >= 4836 && facei <= 4847) || facei == 20253 
            || facei == 29283 || facei == 16290)
        {
            Info << facei << " r = " << r << " r1 = " << r1 
                 << " fCtrs = " << fCtrs[facei] << endl;
        }
*/
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

const vectorField& primitiveMesh::faceCentres() const
{
    if (!faceCentresPtr_)
    {
        calcFaceCentresAndAreas();
    }

    return *faceCentresPtr_;
}


const vectorField& primitiveMesh::faceAreas() const
{
    if (!faceAreasPtr_)
    {
        calcFaceCentresAndAreas();
    }

    return *faceAreasPtr_;
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
